{-------------------------------------------------------------------}
{ DSNODES.INC                                                       }
{ Copyright (c) 1993, 1994  Julian M. Bucknall. All Rights Reserved }
{-------------------------------------------------------------------}

const
  ListInitialised : boolean = false;
  MinNodeSize = 8;
  MaxNodeSize = 32;
  NodeSizeDelta = 4;
  NumNodeSizes = succ((MaxNodeSize - MinNodeSize) div NodeSizeDelta);

var
  NodeStoreList : array [0..pred(NumNodeSizes)] of PNodeStore;

function GetNodeStore(Size : word) : PNodeStore;
  var
    Index : word;
    NS : PNodeStore;
  begin
    GetNodeStore := nil;

    if (Size < MinNodeSize) then
      Size := MinNodeSize
    else if (Size > MaxNodeSize) then
      Size := MaxNodeSize;
    Size := (pred(Size + NodeSizeDelta) div NodeSizeDelta) * NodeSizeDelta;

    if not ListInitialised then
      begin
        FillChar(NodeStoreList, sizeof(NodeStoreList), 0);
        ListInitialised := true;
      end;

    Index := (Size - MinNodeSize) div NodeSizeDelta;

    NS := NodeStoreList[Index];
    if not Assigned(NS) then
      begin
        NS := New(PNodeStore, Init(Size));
        if not Assigned(NS) then
          Exit;
        NodeStoreList[Index] := NS;
      end;
    inc(NS^.RefCount);

    GetNodeStore := NS;
  end;

procedure FreeNodeStore(NS : PNodeStore);
  var
    Index : word;
  begin
    if Assigned(NS) then
      begin
        dec(NS^.RefCount);
        if (NS^.RefCount = 0) then
          begin
            Index := (NS^.NodeSize - MinNodeSize) div NodeSizeDelta;
            Dispose(NS, Done);
            NodeStoreList[Index] := nil;
          end;
      end;
  end;

{=TNodeStore==========================================================
A node warehouse.

A node warehouse stores nodes for TAbstractContainer descendants.
Because the size of a node for a given container is fixed, the
TNodeStore can preallocate a single block of them, and dole them out
singly to the requesting container (ie suballocate the larger block
into smaller nodes). When a node is finished with, it is returned to
the store and will be doled out again. The node store manages two
structures: a very simple linked list of node blocks and a simple
stack of used nodes. The node block is 128 nodes large. This extra
effort is well rewarded, compared with allocating nodes when and
where needed from the heap manager, this is noticeably faster (15-
20% faster, dependent on the number of allocations/frees of nodes).

The node warehouses are stored in a simple array as a global resource.
For each node size there will be one node warehouse. There can be many
containers attached to each warehouse, the count is held in the
RefCount field. Every time a container gets attached to a node
warehouse RefCount is incremented, every time one is unlinked the
RefCount is decremented. If it reaches zero, it is freed. Node
warehouses are allocated with GetNodeStore and freed with
FreeNodeStore.

19Sep93 JMB
======================================================================}
const
  NumNodes = 128; {Best if it is a power of two}
type
  OS = record O, S : word; end;
{--------}
constructor TNodeStore.Init(aNodeSize : word);
  begin
    NodeSize := aNodeSize;
    if not SafeGetMem(Block, aNodeSize * NumNodes) then
      begin
        ConstructorError := edsOutOfMemory;
        Fail;
      end;
    Block^.Link := nil;
    CurOffset := (aNodeSize * NumNodes) - aNodeSize;
    NodeStack := Alloc;
    NodeStack^.Link := NodeStack;
    SpareNodeCount := 0;
    RefCount := 0;
  end;
{--------}
destructor TNodeStore.Done;
  var
    Temp : PNode;
  begin
    repeat
      Temp := Block;
      Block := Temp^.Link;
      SafeFreeMem(Temp, NodeSize * NumNodes);
    until not Assigned(Block);
  end;
{--------}
function TNodeStore.Alloc : PNode;
  var
    Temp,
    Node : PNode;
    DoneIt : boolean;
  begin
    repeat
      DoneIt := true;
      if (CurOffset > 0) then
        begin
          Node := Block;
          inc(OS(Node).O, CurOffset);
          dec(CurOffset, NodeSize);
        end
      else if (SpareNodeCount > 0) then
        begin
          Node := NodeStack^.Link;
          NodeStack^.Link := Node^.Link;
          dec(SpareNodeCount);
        end
      else
        begin
          if not SafeGetMem(Temp, NodeSize * NumNodes) then
            Node := nil
          else
            begin
              Temp^.Link := Block;
              Block := Temp;
              CurOffset := (NodeSize * NumNodes) - NodeSize;
              DoneIt := false;
            end;
        end;
    until DoneIt;
    Alloc := Node;
  end;
{--------}
procedure TNodeStore.Dealloc(aNode : PNode);
  begin
    if Assigned(aNode) then
      begin
        {$IFDEF DEBUG}
        FillChar(aNode^, NodeSize, $CC);
        {$ENDIF}
        aNode^.Link := NodeStack^.Link;
        NodeStack^.Link := aNode;
        inc(SpareNodeCount);
      end;
  end;
{---------------------------------------------------------------------}
