{===EZSTRUCS==========================================================

A data structures unit: defines a stack, queue, deque, several lists,
binary tree, binary search tree, red-black search tree, and priority
queue.

EZStrucs is Copyright (c) 1993, 1994 by  Julian M. Bucknall

VERSION HISTORY
05Sep94 JMB 1.10 consolidation and initial release as freeware
======================================================================}
{ Copyright (c) 1993, 1994, Julian M. Bucknall. All Rights Reserved   }

unit EZStrucs;

{ Undefine this if you don't want debugging info or assertion checks }
{$DEFINE DEBUG}

{ Undefine this if you want the structures to store more than 64K items }
{$DEFINE MaxIs64KItems}

{ Undefine this if you want to unroll the binary tree recursive routines }
{$DEFINE UseTreeRecursion}


{------Common compiler switches---------------------------------------}
{$A+   Word align variables }
{$B-   Short-circuit boolean expressions }
{$F+   Force Far calls }
{$I-   No I/O checking }
{$N+   Allow coprocessor instructions }
{$P+   Open parameters enabled }
{$Q-   No integer overflow checking }
{$R-   No range checking }
{$S-   No stack checking }
{$T-   @ operator is NOT typed }
{$V-   Disable var string checking }
{$X+   Enable extended syntax }
{$IFDEF DEBUG}
{$D+,L+,Y+  Enable debug information }
{$ENDIF}
{---------------------------------------------------------------------}

{------Real mode compiler switches------------------------------------}
{$IFDEF MSDOS}
{$E+   Enable coprocessor emulation }
{$G-   8086 type instructions }
{$O-   Do NOT allow overlays }
{$DEFINE RealMode}
{$UNDEF  ProtMode}
{$ENDIF}
{---------------------------------------------------------------------}

{------Protected mode compiler switches-------------------------------}
{$IFDEF DPMI}
{$E+   Enable coprocessor emulation }
{$G+   80286+ type instructions }
{$UNDEF  RealMode}
{$DEFINE ProtMode}
{$ENDIF}
{---------------------------------------------------------------------}

{------Windows compiler switches--------------------------------------}
{$IFDEF WINDOWS}
{$G+   80286+ type instructions }
{$K+   Use smart callbacks
{$W-   No Windows realmode stack frame }
{$UNDEF  RealMode}
{$DEFINE ProtMode}
{$ENDIF}
{---------------------------------------------------------------------}

interface

uses
  EZRand,
  Objects;                  {Objects are derived from TObject}

var
  ConstructorError : word;  {If an Init fails, this has the error code}

{ EZStrucs error codes }
const
  edsNoError        = 0;  {Success}
  edsOutOfMemory    = 1;  {Memory allocation failed}
  edsTooManyItems   = 2;  {Container has >64K items - MaxIs64KItems is on}
  edsInsInvalidHere = 3;  {Cannot insert at this node}
  edsDelInvalidHere = 4;  {Cannot delete this node}
  edsInsertDup      = 5;  {Attempt to insert duplicate data}
  edsTreeStackError = 6;  {Stack problem with pre/in/postorder traversal - out of memory}
  edsTreeQueueError = 7;  {Queue problem with levelorder traversal - out of memory}
  edsCannotMoveHere = 8;  {Moving outside the container's boundary}
  edsCloneError     = 9;  {Error whilst cloning container: couldn't dup data, out of memory}
  edsIncompatible   =10;  {Attempt to join two containers with different sized nodes}

const
  skMaxLevels = 16;       {Maximum levels in a skip list}

type
  TChild = (CLeft, CRight);
    {-Binary trees: flags for left and right children}
  TTraversalType = (PreOrder, InOrder, PostOrder, LevelOrder);
    {-Binary trees: methods of traversing their nodes}

  TDataItem = record          {A Data item - various typecasts of a pointer}
    case byte of
      0 : (Value : longint);  {A value for byte, ..., longint}
      1 : (Obj   : PObject);  {A pointer to a TObject descendant}
      2 : (St    : PString);  {A pointer to a string}
      3 : (Ch    : char);     {A character}
  end;

  {Forward pointer type declarations}
  PNode = ^TNode;
  PNodeStore = ^TNodeStore;
  PAbstractContainer = ^TAbstractContainer;
  PStack = ^TStack;
  PQueue = ^TQueue;
  PDeque = ^TDeque;
  PLinkList = ^TLinkList;
  PDList = ^TDList;
  PSkipList = ^TSkipList;
  PBinTree = ^TBinTree;
  PBinSearchTree = ^TBinSearchTree;
  PrbSearchTree = ^TrbSearchTree;
  PPriorityQueue = ^TPriorityQueue;


  TListCursor = longint;
    {-Cursor for TDList and TSkipList (double linked & skip lists)}
  TTreeCursor = longint;
    {-Cursor for TBinTree and descendants (binary trees)}

  TDisposeDataProc = procedure (aData : pointer);
    {-Data disposal procedure type for containers}
  TCompareFunc = function (Data1, Data2 : pointer) : integer;
    {-Data comparison procedure type for containers}
  TDupDataFunc = function (aData : pointer; var NewData : pointer) : boolean;
    {-Data duplication procedure type for containers}

  TIteratorProc = procedure (C : PAbstractContainer; aData : pointer;
                             ExtraData : pointer);
    {-ForEach iterator procedure definition}
  TIteratorFunc = function  (C : PAbstractContainer; aData : pointer;
                             ExtraData : pointer) : boolean;
    {-FirstThat iterator function definition}


  {--Internal object type definitions--}
  TNode = record
    {-Internal definition of a node}
    Data : pointer;
    case byte of                           {For...}
      0 : (Link : PNode);                  {Stacks, Queues, Deques, Lists}
      1 : (FLink, BLink : PNode);          {Doubly-linked lists}
      2 : (Size   : word;                  {Skip lists}
           Lvls   : word;
           BkLink : PNode;
           FwLink : array [0..pred(skMaxLevels)] of PNode);
      3 : (TLink : array [TChild] of PNode;{Trees}
           case byte of
             0 : (PKC : longint);          {Binary Trees}
             1 : (PLink : PNode))          {Heaps}
  end;

  TNodeStore = object
    {-Internal object that maintains suballocation of nodes}
    RefCount  : word;
    CurOffset,
    NodeSize  : word;
    Block     : PNode;
    NodeStack : PNode;
    SpareNodeCount : longint;

    constructor Init(aNodeSize : word);
    destructor Done; virtual;

    function  Alloc : PNode;
    procedure Dealloc(aNode : PNode);
  end;


  {--Container object types--}
  TAbstractContainer = object(TObject)
    {-Ancestor object: methods will be overridden}
    public
      {$IFDEF MaxIs64KItems}
      Count    : word;
      {$ELSE}
      Count    : longint;
      {$ENDIF}

   {protected}
      NodeSize : word;
      NS       : PNodeStore;
      InDone   : boolean;
      OwnsData : boolean;

      Compare     : TCompareFunc;
      DisposeData : TDisposeDataProc;
      DupData     : TDupDataFunc;

    public
      constructor Init(aNodeSize : word; DataOwner : boolean);
      destructor Done; virtual;

      function  Clone(DataOwner : boolean; NewCompare : TCompareFunc) : PAbstractContainer; virtual;
      procedure Empty; virtual;
      procedure Error(Code : word); virtual;
      function  IsEmpty : boolean;
      function  NewContainer(DataOwner : boolean) : PAbstractContainer; virtual;
      procedure SetCompare(NewFunc : TCompareFunc);
      procedure SetDisposeData(NewProc : TDisposeDataProc);
      procedure SetDupData(NewFunc : TDupDataFunc);

   {protected}
      procedure acDisposeNode(aNode : PNode); virtual;
      function  acNewNode(aData : pointer) : PNode; virtual;
  end;

  TStack = object(TAbstractContainer)
    {-Stack object}
    private
      Head : PNode;

    public
      constructor Init(DataOwner : boolean);

      function  Clone(DataOwner : boolean; NewCompare : TCompareFunc) : PAbstractContainer; virtual;
      procedure Empty; virtual;
      function  Examine : pointer;
      function  NewContainer(DataOwner : boolean) : PAbstractContainer; virtual;
      function  Pop : pointer;
      function  Push(aData : pointer) : boolean;
  end;

  TQueue = object(TAbstractContainer)
    {-Queue object}
    private
      Head, Tail : PNode;

    public
      constructor Init(DataOwner : boolean);

      function  Append(aData : pointer) : boolean;
      function  Clone(DataOwner : boolean; NewCompare : TCompareFunc) : PAbstractContainer; virtual;
      procedure Empty; virtual;
      function  Examine : pointer;
      function  NewContainer(DataOwner : boolean) : PAbstractContainer; virtual;
      function  Pop : pointer;
  end;

  TDeque = object(TQueue)
    {-Deque object}
    public
      function  NewContainer(DataOwner : boolean) : PAbstractContainer; virtual;
      function  Push(aData : pointer) : boolean;
  end;

  TLinkList = object(TAbstractContainer)
    {-Single linked list object}
    public
      IsSorted: boolean;

    private
      Cursor, BF, AL  : PNode;

    public
      constructor Init(DataOwner : boolean);

      function  Clone(DataOwner : boolean; NewCompare : TCompareFunc) : PAbstractContainer; virtual;
      procedure Delete;
      procedure Empty; virtual;
      procedure Erase;
      function  Examine : pointer;
      function  FirstThat(Action : TIteratorFunc; ExtraData : pointer) : pointer;
      procedure ForEach(Action : TIteratorProc; ExtraData : pointer);
      function  InsertAfter(aData : pointer) : boolean;
      function  InsertBefore(aData : pointer) : boolean;
      function  InsertSorted(aData : pointer) : boolean;
      function  IsAfterLast : boolean;
      function  IsBeforeFirst : boolean;
      procedure Join(List : PLinkList);
      function  NewContainer(DataOwner : boolean) : PAbstractContainer; virtual;
      procedure Next;
      procedure Prev;
      function  Replace(aData : pointer) : pointer;
      function  Search(aData : pointer) : boolean;
      procedure SetBeforeFirst;
      procedure SetAfterLast;
      function  Split : PLinkList;
  end;

  TDList = object(TAbstractContainer)
    {-Double linked list object}
    public
      IsSorted: boolean;

    private
      BF, AL  : PNode;

    public
      constructor Init(DataOwner : boolean);

      function  Clone(DataOwner : boolean; NewCompare : TCompareFunc) : PAbstractContainer; virtual;
      function  Delete(Cursor : TListCursor) : TListCursor;
      procedure Empty; virtual;
      function  Erase(Cursor : TListCursor) : TListCursor;
      function  Examine(Cursor : TListCursor) : pointer;
      function  FirstThat(Action : TIteratorFunc; ExtraData : pointer) : TListCursor;
      procedure ForEach(Action : TIteratorProc; ExtraData : pointer);
      function  InsertAfter(Cursor : TListCursor; aData : pointer) : boolean;
      function  InsertBefore(Cursor : TListCursor; aData : pointer) : boolean;
      function  InsertSorted(aData : pointer) : boolean;
      function  IsAfterLast(Cursor : TListCursor) : boolean;
      function  IsBeforeFirst(Cursor : TListCursor) : boolean;
      procedure Join(Cursor : TListCursor; List : PDList);
      function  NewContainer(DataOwner : boolean) : PAbstractContainer; virtual;
      function  Next(Cursor : TListCursor) : TListCursor;
      function  Prev(Cursor : TListCursor) : TListCursor;
      function  Replace(Cursor : TListCursor; aData : pointer) : pointer;
      function  Search(var Cursor : TListCursor; aData : pointer) : boolean;
      function  SetBeforeFirst : TListCursor;
      function  SetAfterLast : TListCursor;
      function  Split(Cursor : TListCursor) : PDList;
  end;

  TSkipList = object(TAbstractContainer)
    {-Skip linked list object}
    private
      BF, AL    : PNode;
      CurLevels : integer;
      skNewNodeLevel : integer;

    public
      constructor Init(DataOwner : boolean);

      function  Clone(DataOwner : boolean; NewCompare : TCompareFunc) : PAbstractContainer; virtual;
      function  Delete(Cursor : TListCursor) : TListCursor;
      procedure Empty; virtual;
      function  Erase(Cursor : TListCursor) : TListCursor;
      function  Examine(Cursor : TListCursor) : pointer;
      function  FirstThat(Action : TIteratorFunc; ExtraData : pointer) : TListCursor;
      procedure ForEach(Action : TIteratorProc; ExtraData : pointer);
      function  Insert(var Cursor : TListCursor; aData : pointer) : boolean;
      function  IsAfterLast(Cursor : TListCursor) : boolean;
      function  IsBeforeFirst(Cursor : TListCursor) : boolean;
      procedure Join(List : PSkipList);
      function  NewContainer(DataOwner : boolean) : PAbstractContainer; virtual;
      function  Next(Cursor : TListCursor) : TListCursor;
      function  Prev(Cursor : TListCursor) : TListCursor;
      function  Replace(Cursor : TListCursor; aData : pointer) : pointer;
      function  Search(var Cursor : TListCursor; aData : pointer) : boolean;
      function  SetBeforeFirst : TListCursor;
      function  SetAfterLast : TListCursor;
      function  Split(Cursor : TListCursor) : PSkipList;

   {protected}
      procedure acDisposeNode(aNode : PNode); virtual;
      function  acNewNode(aData : pointer) : PNode; virtual;
  end;

  TBinTree = object(TAbstractContainer)
    {-Binary tree object}
    private
      NilNode,
      Rt       : PNode;
      TravType : TTraversalType;

    public
      constructor Init(DataOwner : boolean);

      function  Clone(DataOwner : boolean; NewCompare : TCompareFunc) : PAbstractContainer; virtual;
      function  Delete(Cursor : TTreeCursor) : TTreeCursor; virtual;
      procedure Empty; virtual;
      function  Erase(Cursor : TTreeCursor) : TTreeCursor;
      function  Examine  (Cursor : TTreeCursor) : pointer;
      function  FirstThat(Action : TIteratorFunc; ExtraData : pointer) : TTreeCursor;
      procedure ForEach  (Action : TIteratorProc; ExtraData : pointer);
      function  Insert   (var Cursor : TTreeCursor; aData : pointer) : boolean; virtual;
      function  IsLeaf   (Cursor : TTreeCursor) : boolean;
      function  IsRoot   (Cursor : TTreeCursor) : boolean;
      procedure Join(Cursor : TTreeCursor; Tree : PBinTree); virtual;
      function  Left(Cursor : TTreeCursor) : TTreeCursor;
      function  NewContainer(DataOwner : boolean) : PAbstractContainer; virtual;
      function  Parent(Cursor : TTreeCursor) : TTreeCursor;
      function  Replace  (Cursor : TTreeCursor; aData : pointer) : pointer;
      function  Right(Cursor : TTreeCursor) : TTreeCursor;
      function  Root : TTreeCursor;
      function  Search   (var Cursor : TTreeCursor; aData : pointer) : boolean; virtual;
      procedure SetTraversalType(TT : TTraversalType);
   (* function  Split(Cursor : TTreeCursor) : PBinTree; virtual; **not implemented**)

   {protected}
      function  acNewNode  (aData : pointer) : PNode; virtual;
  end;

  TBinSearchTree = object(TBinTree)
    {-Binary search tree object}
    public
      function  Clone(DataOwner : boolean; NewCompare : TCompareFunc) : PAbstractContainer; virtual;
      function  Delete (Cursor : TTreeCursor) : TTreeCursor;  virtual;
      function  Insert (var Cursor : TTreeCursor; aData : pointer) : boolean; virtual;
      procedure Join(Cursor : TTreeCursor; Tree : PBinTree); virtual;
      function  NewContainer(DataOwner : boolean) : PAbstractContainer; virtual;
      function  Replace(Cursor : TTreeCursor; aData : pointer) : pointer;
      function  Search (var Cursor : TTreeCursor; aData : pointer) : boolean; virtual;
   (* function  Split(Cursor : TTreeCursor) : PBinTree; virtual; **not implemented**)

   {protected}
      procedure bsSwapData(OldCursor, NewCursor : TTreeCursor); virtual;
  end;

  TrbSearchTree = object(TBinSearchTree)
    {-Balanced binary search tree object (Red-black tree)}
    private
      DeletedNodeWasBlack : boolean;

    public
      function  Delete (Cursor : TTreeCursor) : TTreeCursor;  virtual;
      function  Insert (var Cursor : TTreeCursor; aData : pointer) : boolean; virtual;
      function  NewContainer(DataOwner : boolean) : PAbstractContainer; virtual;

   {protected}
      procedure bsSwapData(OldCursor, NewCursor : TTreeCursor); virtual;
      function  rbPromote(Cursor : TTreeCursor) : TTreeCursor;
  end;

  TPriorityQueue = object(TAbstractContainer)
    {-Priority queue object}
    private
      Rt : PNode;

    public
      constructor Init(DataOwner : boolean);

      function  Append(aData : pointer) : boolean;
      function  Clone(DataOwner : boolean; NewCompare : TCompareFunc) : PAbstractContainer; virtual;
      procedure Empty; virtual;
      function  Examine : pointer;
      function  NewContainer(DataOwner : boolean) : PAbstractContainer; virtual;
      function  Pop : pointer;
      function  Replace(aData : pointer) : pointer;

   {protected}
      procedure pqBubbleUp(Node : PNode);
      function  pqGetNodeFromIndex(Inx : longint) : PNode;
      procedure pqSwapData(Node1, Node2 : PNode);
      procedure pqTrickleDown(Node : PNode);
  end;

{---String data object routines---}
function  EZStrCompare(Data1, Data2 : pointer) : integer;
procedure EZStrDisposeData(aData : pointer);
function  EZStrDupData(aData : pointer; var NewData : pointer) : boolean;

{---Longint data object routines---}
function  EZIntCompare(Data1, Data2 : pointer) : integer;
procedure EZIntDisposeData(aData : pointer);
function  EZIntDupData(aData : pointer; var NewData : pointer) : boolean;

{---Do nothing comparison routine---}
function EZNoCompare(Data1, Data2 : pointer) : integer;

implementation

{---Data object routines----------------------------------------------}
function  EZAbstractCompare(Data1, Data2 : pointer) : integer;
  begin
    Abstract;
  end;
{--------}
procedure EZAbstractDisposeData(aData : pointer);
  begin
    {do nothing}
  end;
{--------}
function  EZAbstractDupData(aData : pointer; var NewData : pointer) : boolean;
  begin
    Abstract;
  end;
{--------}
function  EZStrCompare(Data1, Data2 : pointer) : integer;
  begin
    if (Data1 = nil) then
      if (Data2 = nil) then
           EZStrCompare := 0
      else EZStrCompare := -1
    else
      if (Data2 = nil) then EZStrCompare := 1
      else
        asm
          xor ax, ax
          mov cx, ax
          mov dx, ds
          lds si, Data1
          les di, Data2
          mov bl, [si]
          inc si
          mov bh, es:[di]
          inc di
          mov cl, bl
          cmp cl, bh
          jbe @@CompareStrings
          mov cl, bh
        @@CompareStrings:
          or cx, cx
          jz @@CompareLengths
          cld
          repe cmpsb
          jb @@LT
          ja @@GT
        @@CompareLengths:
          cmp bl, bh
          je @@Exit
          jb @@LT
        @@GT:
          inc ax
          inc ax
        @@LT:
          dec ax
        @@Exit:
          mov @Result, ax
          mov ds, dx
        end;
  end;
{--------}
procedure EZStrDisposeData(aData : pointer);
  begin
    DisposeStr(PString(aData));
  end;
{--------}
function  EZStrDupData(aData : pointer; var NewData : pointer) : boolean;
  begin
    EZStrDupData := true;
    if (aData = nil) then
      NewData := nil
    else
      begin
        NewData := NewStr(PString(aData)^);
        if (NewData = nil) then
          EZStrDupData := false;
      end;
  end;
{--------}
function  EZIntCompare(Data1, Data2 : pointer) : integer; assembler;
  asm
    xor ax, ax                  {assume equal}
    mov bx, Data1.Word[2]       {get high integer of Data1}
    cmp bx, Data2.Word[2]       {compare with high integer of Data2}
    jl @@LT                     {less than? jump if so}
    jg @@GT                     {greater than? jump if so}
    mov bx, Data1.Word[0]       {get low word of Data1}
    cmp bx, Data2.Word[0]       {compare with low word of Data2}
    je @@Exit                   {equal? exit if so}
    jb @@LT                     {less than? jump if so}
  @@GT:                         {get here if Data1 > Data2}
    inc ax                      {set ax to 2}
    inc ax                      {fall through next bit to set ax to 1}
  @@LT:                         {get here if Data1 < Data2}
    dec ax                      {decrement ax by 1}
  @@Exit:
  end;
{--------}
procedure EZIntDisposeData(aData : pointer);
  begin
    {do nothing}
  end;
{--------}
function  EZIntDupData(aData : pointer; var NewData : pointer) : boolean;
  begin
    NewData := aData;
    EZIntDupData := true;
  end;
{--------}
function EZNoCompare(Data1, Data2 : pointer) : integer; assembler;
  asm
    xor ax, ax {return 0}
  end;
{---------------------------------------------------------------------}


{$IFDEF DEBUG}
{=Assert==============================================================
If the proposition is false, aborts the program. For DEBUG modes only.
05Sep94 JMB
======================================================================}
procedure Assert(Proposition : boolean; const Where : string);
  var
    TempZ : array [0..255] of char;
  begin
    if (Proposition = false) then
      begin
        {$IFDEF Windows}
        MessageBox(0, StrPCopy(TempZ, Where), 'Assertion Error', MB_OK);
        {$ELSE}
        writeln;
        writeln;
        writeln('Assertion Error: ', Where);
        {$ENDIF}
        Halt(255);
      end;
  end;

{=HaltOnError=========================================================
Aborts the program with the passed error message. For DEBUG modes
only.
05Sep94 JMB
======================================================================}
procedure HaltOnError(const ErrorMsg : string);
  var
    TempZ : array [0..255] of char;
  begin
    {$IFDEF Windows}
    MessageBox(0, StrPCopy(TempZ, ErrorMsg), 'Error', MB_OK);
    {$ELSE}
    writeln;
    writeln;
    writeln('Error: ', ErrorMsg);
    {$ENDIF}
    Halt(255);
  end;
{$ENDIF}

{===SafeFreeMem/SafeGetMem===========================================
Allocates and deallocates memory 'safely'. In debug mode SafeGetMem
will fill the allocated block with $CC characters - in the debugger
it will show up data in the block that wasn't initialised, and if
there is a bug such that the block is executed as code you'll get an
automatic breakpoint ($CC = INT $03).
05Sep94 JMB
=====================================================================}
function  SafeGetMem(var P; BlockSize : word) : boolean;
  const
    HeapLeeway = 4096;     {SafeGetMem won't alloc into this leeway}
  var
    Pt : pointer absolute P;
  begin
    if (BlockSize <= MaxAvail+HeapLeeway) then
      begin
        GetMem(Pt, BlockSize);
        {$IFDEF DEBUG}
        FillChar(Pt^, BlockSize, $CC);
        {$ENDIF}
        SafeGetMem := Assigned(Pt);
      end
    else
      Pt := nil;
  end;
{--------}
procedure SafeFreeMem(var P; BlockSize : word);
  var
    Pt : pointer absolute P;
  begin
    if Assigned(Pt) then
      begin
        FreeMem(Pt, BlockSize);
        Pt := nil;
      end;
  end;


{$I DSNodes.INC}     {TNodeStore}
{$I DSAbsCtr.INC}    {TAbstractContainer}
{$I DSStack.INC}     {TStack}
{$I DSQueue.INC}     {TQueue and TDeque}
{$I DSList.INC}      {TLinkList}
{$I DSDList.INC}     {TDList}
{$I DSSkpLst.INC}    {TSkipList}
{$I DSBiTree.INC}    {TBinTree}
{$I DSBSTree.INC}    {TBinSearchTree}
{$I DSrbTree.INC}    {TrbSearchTree}
{$I DSPriorQ.INC}    {TPriorityQueue}


end.
